import matplotlib.pyplot as plt
import matplotlib.collections
import matplotlib.patches
import numpy as np
from matplotlib import rc
from colours import *
from errortools import *
import mpl_axes_aligner


rc('text', usetex=True)
rc('text.latex', preamble="\n".join([
    '\\usepackage{amsmath}',
    '\\newcommand{\\D}{\mathrm{d}}'
]))


def watermark(fig, txt='PRELIMINARY', fontsize=60, rotation=20):
    axs = fig.axes
    corners = matplotlib.transforms.Bbox.union(
        [i.get_position() for i in axs]
    ).corners()
    center = np.average(corners, axis=0)
    fig.text(
        center[0], center[1],
        txt,
        fontsize=fontsize, rotation=rotation,
        ha='center', va='center', alpha=0.1
    )


def errorband(p, ax=None, col='default', underflow=False, overflow=False):
    if ax is None:
        ax = plt.gca()

    p = p.copy()

    delta = p[2,0] - p[1,0]

    if p[0,0] == -np.inf:
        if underflow is False:
            p = p[1:]
        else:
            s = 2 if type(underflow) == bool else underflow
            p[0,0] = p[1,0] - (s+0.5) * delta
    if p[-1,0] == +np.inf:
        if overflow is False:
            p = p[:-1]
        else:
            s = 2 if type(overflow) == bool else overflow
            p[-1,0] = p[-2,0] + (s+0.5) * delta

    boxes = [
        matplotlib.patches.Rectangle(
            (x - delta, y - e/2), delta, e
        )
        for x, y, e in p
    ]

    if col == 'default':
        artist = ax.step(p[:,0], p[:,1])
        col = artist[0].get_color()
    else:
        artist = ax.step(p[:,0], p[:,1], col)

    pc = matplotlib.collections.PatchCollection(
        boxes,
        facecolor=col, edgecolor=col
    )

    ax.add_collection(pc)
    return artist


def twopanel(labelx,
             upleft=[], labupleft="", colupleft=defcol,
             downleft=[], labdownleft="", coldownleft=defcol,
             upright=[], labupright="", colupright=defcol,
             downright=[], labdownright="", coldownright=defcol,
             upalign=[], downalign=[]):

    if type(upleft) == np.ndarray:
        upleft = [upleft]
    if type(upright) == np.ndarray:
        upright = [upright]
    if type(downleft) == np.ndarray:
        downleft = [downleft]
    if type(downright) == np.ndarray:
        downright = [downright]

    fig, axs = plt.subplots(
        2, sharex=True, sharey=True, gridspec_kw={'hspace': 0}
    )
    axs[1].set_xlabel(labelx)

    axs[0].set_ylabel(labupleft)
    for i, c in zip(upleft, colupleft):
        errorband(i, ax=axs[0], col=c)

    if len(upright) > 0:
        ax2 = axs[0].twinx()
        if labupright is not None:
            ax2.set_ylabel(labupright)
        for i, c in zip(upright, colupright):
            errorband(i, ax=ax2, col=c)
        if len(upalign) == 2:
            mpl_axes_aligner.yaxes(
                axs[0], upalign[0],
                ax2, upalign[1]
            )

    axs[1].set_ylabel(labdownleft)
    for i, c in zip(downleft, coldownleft):
        errorband(i, ax=axs[1], col=c)

    if len(downright) > 0:
        ax3 = axs[1].twinx()
        if labdownright is not None:
            ax3.set_ylabel(labdownright)
        for i, c in zip(downright, coldownright):
            errorband(i, ax=ax3, col=c)
        if len(downalign) == 2:
            mpl_axes_aligner.yaxes(
                axs[1], downalign[0],
                ax3, downalign[1]
            )

    return fig, axs


def kplot(sigma, labelx='x_e', labelsigma=None,
          labelknlo="$K^{(1)}$", labelknnlo="$K^{(2)}$",
          show=[0,-1], showk=[1,2]):

    if labelsigma is None:
        labelsigma = r"$\D\sigma\,/\,\D " + labelx + "$"

    kwargs = {
        "labelx":"$%s$" % labelx,
        "labupleft": labelsigma,
        "downalign": [1,1]
    }

    if type(sigma) != dict:
        raise ValueError("kplot takes dicts!")
    if "lo" not in sigma:
        raise KeyError("kplot: no leading order")
    if "nlo" not in sigma:
        raise KeyError("kplot: no next-to-leading order")
    if "nnlo" in sigma:
        orders = ['lo', 'nlo', 'nnlo']
        xsec = {
            'lo': sigma['lo'],
            'nlo': addplots(sigma['lo'], sigma['nlo']),
            'nnlo': addplots(
                sigma['lo'],
                addplots(sigma['nlo'], sigma['nnlo'])
            )
        }
    else:
        orders = ['lo', 'nlo']
        xsec = {
            'lo': sigma['lo'],
            'nlo': addplots(sigma['lo'], sigma['nlo'])
        }

    one = np.zeros(xsec['lo'].shape)
    one[:,1] = 1

    if 1 in showk:
        kwargs['downleft'] = divideplots(sigma['nlo'], sigma['lo']) + one
        kwargs['coldownleft'] = [orderscheme['nlo']]
    if 2 in showk and 'nnlo' in sigma:
        kwargs['downright'] = divideplots(sigma['nnlo'], sigma['lo']) + one
        kwargs['coldownright'] = [orderscheme['nnlo']]

    kwargs['upleft'] = [
        xsec[orders[i]] for i in show
    ]
    kwargs['colupleft'] = [
        orderscheme[orders[i]] for i in show
    ]

    fig, axs = twopanel(**kwargs)

    axs[1].axhline(1, color='black', linewidth=1, zorder=1)

    return fig, axs
